from pathlib import Path

from .parser import OpCode


def trace(code: bytearray, stack: list[int], registers: dict[int, int], memory: list[int]) -> None:
    output = "--- CODE ---\n"

    for i in range(len(code) // 16 + 1):
        output += " ".join([hex(ins)[2:].zfill(2) for ins in code[i * 16 : (i + 1) * 16]]) + "\n"

    output += "\n--- STACK ---\n"

    for item in stack:
        output += str(item) + "\n"

    output += "\n--- REGISTERS ---\n"

    for key, value in registers.items():
        output += f"{chr(96 + key)}: {value}\n"

    output += "\n--- MEMORY ---\n"

    for i in range(len(memory) // 16 + 1):
        output += " ".join([hex(ins)[2:].zfill(2) for ins in memory[i * 16 : (i + 1) * 16]]) + "\n"

    Path("semby.trace").write_text(output)

def execute(code: bytearray, memsize: int = 256) -> None:
    stack: list[int] = []
    registers: dict[int, int] = {}
    memory: list[int] = [0] * memsize

    ptr = 0

    while ptr < len(code):
        instr = code[ptr]

        if instr == OpCode.HLT:
            exit(stack.pop())
        elif instr == OpCode.LD:
            stack.append(registers[code[ptr + 1]])
            ptr += 2
        elif instr == OpCode.ST:
            registers[code[ptr + 1]] = stack.pop()
            ptr += 2
        elif instr == OpCode.STC:
            registers[code[ptr + 1]] = int.from_bytes(code[ptr + 2 : ptr + 10], "little")
            ptr += 10
        elif instr == OpCode.DUP:
            stack.append(stack[-1])
            ptr += 1
        elif instr == OpCode.PUSH:
            stack.append(int.from_bytes(code[ptr + 1 : ptr + 9], "little"))
            ptr += 9
        elif instr == OpCode.JMP:
            ptr = int.from_bytes(code[ptr + 1 : ptr + 5], "little")
        elif instr == OpCode.JMPZ:
            ptr = int.from_bytes(code[ptr + 1 : ptr + 5], "little") if stack.pop() == 0 else ptr + 5
        elif instr == OpCode.JMPNZ:
            ptr = int.from_bytes(code[ptr + 1 : ptr + 5], "little") if stack.pop() != 0 else ptr + 5
        elif instr == OpCode.JMPP:
            ptr = int.from_bytes(code[ptr + 1 : ptr + 5], "little") if stack.pop() > 0 else ptr + 5
        elif instr == OpCode.ADD:
            stack.append(stack.pop() + stack.pop())
            ptr += 1
        elif instr == OpCode.SUB:
            stack.append(stack.pop() - stack.pop())
            ptr += 1
        elif instr == OpCode.DIV:
            stack.append(int(stack.pop() // stack.pop()))
            ptr += 1
        elif instr == OpCode.MUL:
            stack.append(stack.pop() * stack.pop())
            ptr += 1
        elif instr == OpCode.MOD:
            stack.append(stack.pop() % stack.pop())
            ptr += 1
        elif instr == OpCode.OUT:
            print(stack.pop())
            ptr += 1
        elif instr == OpCode.OUTC:
            print(chr(stack.pop()), end="")
            ptr += 1
        elif instr == OpCode.MDP:
            loc = stack.pop()
            val = stack.pop()
            memory[loc] = val
            ptr += 1
        elif instr == OpCode.MLD:
            stack.append(memory[stack.pop()])
            ptr += 1
        elif instr == OpCode.CPY:
            loc_from = stack.pop()
            loc_to = stack.pop()
            memory[loc_to] = memory[loc_from]
            ptr += 1
        elif instr == OpCode.TRC:
            trace(code, stack, registers, memory)
            ptr += 1
        else:
            print("Invalid opcode: ", instr)
            exit(1)



from pathlib import Path

from typer import Typer, echo

from ..core import compile as compile_bc
from ..core import execute, parse

app = Typer()


@app.command(name="compile")
def compile(file: str, out: str = "out.smbc") -> None:
    data = Path(file).read_text()

    compiled = compile_bc(parse(data, file))

    with Path(out).open("wb") as f:
        f.write(compiled)

    echo(f"Compilation done. Written {len(compiled)} bytes to {out}")


@app.command(name="exec")
def exec(file: str, memsize: int = 256) -> None:
    data = Path(file).read_bytes()

    execute(bytearray(data), memsize)


@app.command(name="run")
def run(file: str, memsize: int = 256) -> None:
    data = Path(file).read_text()

    compiled = compile_bc(parse(data, file))

    execute(compiled, memsize)



from dataclasses import dataclass
from re import compile

IDENTIFIER = compile(r"^[a-z]+$")


@dataclass
class Token:
    type: str
    value: str
    line: int
    file: str


def tokenise(code: str, file: str) -> list[list[Token]]:
    lines: list[list[Token]] = []

    for i, line in enumerate(code.split("\n")):
        line = line.split()
        out_line: list[Token] = []

        for part in line:
            if part.startswith(";"):
                break

            if part in ".:":
                out_line.append(Token("sym", part, i + 1, file))
            elif IDENTIFIER.match(part):
                out_line.append(Token("id", part, i + 1, file))
            elif part.isdigit():
                out_line.append(Token("num", part, i + 1, file))
            else:
                raise SyntaxError(f"[{file}:{i+i}] Unknown token: " + part)

        if out_line:
            lines.append(out_line)

    return lines



from dataclasses import dataclass
from enum import IntEnum
from typing import NoReturn

from .lexer import Token, tokenise

REGISTERS = ["a", "b", "c", "d"]
LOADS = [f"ld{r}" for r in REGISTERS]
STORES = [f"st{r}" for r in REGISTERS]


class OpCode(IntEnum):
    HLT = 0x00
    LD = 0x01
    ST = 0x02
    STC = 0x03
    DUP = 0x04
    PUSH = 0x05

    JMP = 0x10
    JMPZ = 0x11
    JMPNZ = 0x12
    JMPP = 0x13

    ADD = 0x20
    SUB = 0x21
    DIV = 0x22
    MUL = 0x23
    MOD = 0x24

    OUT = 0x30
    OUTC = 0x31

    MDP = 0x40
    MLD = 0x41
    CPY = 0x42

    TRC = 0xF0


@dataclass
class Instruction:
    op: int
    data: bytearray


@dataclass
class Jump:
    op: int
    ref: str


@dataclass
class ParserOptions:
    strict_jumps: bool = True


SINGLE = {
    "dup": OpCode.DUP,
    "add": OpCode.ADD,
    "sub": OpCode.SUB,
    "div": OpCode.DIV,
    "mul": OpCode.MUL,
    "mod": OpCode.MOD,
    "out": OpCode.OUT,
    "outc": OpCode.OUTC,
    "mdp": OpCode.MDP,
    "mld": OpCode.MLD,
    "cpy": OpCode.CPY,
    "trc": OpCode.TRC,
}

JUMP = {
    "jmp": OpCode.JMP,
    "jmpz": OpCode.JMPZ,
    "jmpnz": OpCode.JMPNZ,
    "jmpp": OpCode.JMPP,
}


def fail(reason: str, file: str, line: int) -> NoReturn:
    print(f"[{file}:{line}] {reason}")
    exit(1)


def parse(code: str, file: str, options: ParserOptions = None) -> list[Instruction]:
    instructions: list[Instruction | Jump] = []
    jumps: dict[str, int] = {}
    bcoffset: int = 0

    options = options or ParserOptions()

    tokens = tokenise(code, file)

    for line in tokens:
        match line:
            case (Token(type="sym", value=":"), Token(type="id", value=x)):
                if options.strict_jumps and x in jumps:
                    fail(f"Jump location {x} is defined multiple times.", line[0].file, line[0].line)

                jumps[x] = bcoffset
            case (Token(type="id", value="hlt"),):
                instructions.append(Instruction(op=OpCode.HLT, data=bytearray()))
            case (Token(type="id", value=x),) if x in LOADS:
                instructions.append(Instruction(op=OpCode.LD, data=bytearray([LOADS.index(x) + 1])))
                bcoffset += 2
            case (Token(type="id", value=x),) if x in STORES:
                instructions.append(Instruction(op=OpCode.ST, data=bytearray([STORES.index(x) + 1])))
                bcoffset += 2
            case (Token(type="id", value=x), Token(type="num", value=y)) if x in STORES:
                instructions.append(
                    Instruction(
                        op=OpCode.STC, data=bytearray([STORES.index(x) + 1]) + bytearray(int(y).to_bytes(8, "little"))
                    )
                )
                bcoffset += 10
            case (Token(type="id", value="push"), Token(type="num", value=y)):
                instructions.append(Instruction(op=OpCode.PUSH, data=bytearray(int(y).to_bytes(8, "little"))))
                bcoffset += 9
            case (Token(type="id", value=x), Token(type="id", value=y)) if x in JUMP:
                instructions.append(Jump(op=JUMP[x], ref=y))
                bcoffset += 5
            case (Token(type="id", value=x),) if x in SINGLE:
                instructions.append(Instruction(op=SINGLE[x], data=bytearray()))
                bcoffset += 1

    output = []

    for instruction in instructions:
        if isinstance(instruction, Instruction):
            output.append(instruction)
        else:
            offset = jumps[instruction.ref]
            output.append(Instruction(op=instruction.op, data=bytearray(offset.to_bytes(4, "little"))))

    return output


def compile(instructions: list[Instruction]) -> bytearray:
    output = bytearray()

    for instruction in instructions:
        output += instruction.op.to_bytes(1, "little")
        output += instruction.data

    return output



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['semby', 'semby.cli', 'semby.core']

package_data = \
{'': ['*']}

install_requires = \
['typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['semby = semby.cli:app']}

setup_kwargs = {
    'name': 'semby',
    'version': '1.0.3',
    'description': 'A simple stack based assembly like programming language',
    'long_description': '# semby\n\nA simple stack based assembly like programming language\n\n## Installation\n\n`pip install semby`\n\n## Usage\n\n`semby compile <file> && semby exec <file>` or `semby run <file>`\n',
    'author': 'vcokltfre',
    'author_email': 'vcokltfre@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vcokltfre/semby',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
